@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -33,6 +33,7 @@
 #define CONNMAN_API_SUBJECT_TO_CHANGE
 #include <connman/plugin.h>
 #include <connman/resolver.h>
+#include <connman/notifier.h>
 #include <connman/log.h>
 
 #include <glib.h>
@@ -79,6 +80,7 @@ struct server_data {
 	char *server;
 	GIOChannel *channel;
 	guint watch;
+	gboolean enabled;
 };
 
 struct request_data {
@@ -149,7 +151,7 @@ static gboolean server_event(GIOChannel *channel, GIOCondition condition,
 {
 	struct server_data *data = user_data;
 	struct request_data *req;
-	unsigned char buf[768];
+	unsigned char buf[4096];
 	struct domain_hdr *hdr = (void *) &buf;
 	int sk, err, len;
 
@@ -224,12 +226,15 @@ static struct server_data *create_server(const char *interface,
 		return NULL;
 	}
 
-	if (setsockopt(sk, SOL_SOCKET, SO_BINDTODEVICE,
+	if (interface != NULL) {
+		if (setsockopt(sk, SOL_SOCKET, SO_BINDTODEVICE,
 				interface, strlen(interface) + 1) < 0) {
-		connman_error("Failed to bind server %s to interface %s",
+			connman_error("Failed to bind server %s "
+						"to interface %s",
 							server, interface);
-		close(sk);
-		return NULL;
+			close(sk);
+			return NULL;
+		}
 	}
 
 	memset(&sin, 0, sizeof(sin));
@@ -261,12 +266,17 @@ static struct server_data *create_server(const char *interface,
 	g_io_channel_set_close_on_unref(data->channel, TRUE);
 
 	data->watch = g_io_add_watch(data->channel, G_IO_IN,
-							server_event, data);
+						server_event, data);
 
 	data->interface = g_strdup(interface);
 	data->domain = g_strdup(domain);
 	data->server = g_strdup(server);
 
+	/* Enable new servers by default */
+	data->enabled = TRUE;
+
+	connman_info("Adding DNS server %s", data->server);
+
 	return data;
 }
 
@@ -279,6 +289,8 @@ static void destroy_server(struct server_data *data)
 
 	g_io_channel_unref(data->channel);
 
+	connman_info("Removing DNS server %s", data->server);
+
 	g_free(data->server);
 	g_free(data->domain);
 	g_free(data->interface);
@@ -332,41 +344,98 @@ static struct connman_resolver dnsproxy_resolver = {
 	.remove		= dnsproxy_remove,
 };
 
+static void dnsproxy_offline_mode(connman_bool_t enabled)
+{
+	GSList *list;
+
+	DBG("enabled %d", enabled);
+
+	for (list = server_list; list; list = list->next) {
+		struct server_data *data = list->data;
+
+		if (enabled == FALSE) {
+			connman_info("Enabling DNS server %s", data->server);
+			data->enabled = TRUE;
+		} else {
+			connman_info("Disabling DNS server %s", data->server);
+			data->enabled = FALSE;
+		}
+	}
+}
+
+static void dnsproxy_default_changed(struct connman_service *service)
+{
+	GSList *list;
+	char *interface;
+
+	DBG("service %p", service);
+
+	if (service == NULL) {
+		/* When no services are active, then disable DNS proxying */
+		dnsproxy_offline_mode(TRUE);
+		return;
+	}
+
+	interface = connman_service_get_interface(service);
+	if (interface == NULL)
+		return;
+
+	for (list = server_list; list; list = list->next) {
+		struct server_data *data = list->data;
+
+		if (g_strcmp0(data->interface, interface) == 0) {
+			connman_info("Enabling DNS server %s", data->server);
+			data->enabled = TRUE;
+		} else {
+			connman_info("Disabling DNS server %s", data->server);
+			data->enabled = FALSE;
+		}
+	}
+
+	g_free(interface);
+}
+
+static struct connman_notifier dnsproxy_notifier = {
+	.name			= "dnsproxy",
+	.default_changed	= dnsproxy_default_changed,
+	.offline_mode		= dnsproxy_offline_mode,
+};
+
+static unsigned char opt_edns0_type[2] = { 0x00, 0x29 };
+
 static int parse_request(unsigned char *buf, int len,
 					char *name, unsigned int size)
 {
 	struct domain_hdr *hdr = (void *) buf;
 	uint16_t qdcount = ntohs(hdr->qdcount);
+	uint16_t arcount = ntohs(hdr->arcount);
 	unsigned char *ptr;
 	char *last_label = NULL;
-	int label_count = 0;
 	unsigned int remain, used = 0;
 
 	if (len < 12)
 		return -EINVAL;
 
-	DBG("id 0x%04x qr %d opcode %d qdcount %d",
-				hdr->id, hdr->qr, hdr->opcode, qdcount);
+	DBG("id 0x%04x qr %d opcode %d qdcount %d arcount %d",
+					hdr->id, hdr->qr, hdr->opcode,
+							qdcount, arcount);
 
 	if (hdr->qr != 0 || qdcount != 1)
 		return -EINVAL;
 
 	memset(name, 0, size);
 
-	ptr = buf + 12;
-	remain = len - 12;
+	ptr = buf + sizeof(struct domain_hdr);
+	remain = len - sizeof(struct domain_hdr);
 
 	while (remain > 0) {
 		uint8_t len = *ptr;
 
 		if (len == 0x00) {
-			if (label_count > 0)
-				last_label = (char *) (ptr + 1);
+			last_label = (char *) (ptr + 1);
 			break;
 		}
 
-		label_count++;
-
 		if (used + len + 1 > size)
 			return -ENOBUFS;
 
@@ -379,7 +448,29 @@ static int parse_request(unsigned char *buf, int len,
 		remain -= len + 1;
 	}
 
-	DBG("query %s (%d labels)", name, label_count);
+	if (last_label && arcount && remain >= 9 && last_label[4] == 0 &&
+				!memcmp(last_label + 5, opt_edns0_type, 2)) {
+		uint16_t edns0_bufsize;
+
+		edns0_bufsize = last_label[7] << 8 | last_label[8];
+
+		DBG("EDNS0 buffer size %u", edns0_bufsize);
+
+		/* This is an evil hack until full TCP support has been
+		 * implemented.
+		 *
+		 * Somtimes the EDNS0 request gets send with a too-small
+		 * buffer size. Since glibc doesn't seem to crash when it
+		 * gets a response biffer then it requested, just bump
+		 * the buffer size up to 4KiB.
+		 */
+		if (edns0_bufsize < 0x1000) {
+			last_label[7] = 0x10;
+			last_label[8] = 0x00;
+		}
+	}
+
+	DBG("query %s", name);
 
 	return 0;
 }
@@ -542,7 +633,11 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 	for (list = server_list; list; list = list->next) {
 		struct server_data *data = list->data;
 
-		DBG("server %s domain %s", data->server, data->domain);
+		DBG("server %s domain %s enabled %d",
+				data->server, data->domain, data->enabled);
+
+		if (data->enabled == FALSE)
+			continue;
 
 		sk = g_io_channel_unix_get_fd(data->channel);
 
@@ -553,7 +648,11 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 		if (data->domain != NULL) {
 			unsigned char alt[1024];
 			struct domain_hdr *hdr = (void *) &alt;
-			int altlen;
+			int altlen, domlen;
+
+			domlen = strlen(data->domain) + 1;
+			if (domlen < 5)
+				continue;
 
 			alt[0] = req->altid & 0xff;
 			alt[1] = req->altid >> 8;
@@ -566,12 +665,12 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 			if (altlen < 0)
 				continue;
 
-			alt[altlen + 12] = 0x00;
-			alt[altlen + 13] = 0x01;
-			alt[altlen + 14] = 0x00;
-			alt[altlen + 15] = 0x01;
+			altlen += 12;
 
-			err = send(sk, alt, altlen + 12 + 4, 0);
+			memcpy(alt + altlen, buf + altlen - domlen,
+							len - altlen + domlen);
+
+			err = send(sk, alt, len + domlen + 1, 0);
 
 			req->numserv++;
 		}
@@ -671,16 +770,30 @@ static int dnsproxy_init(void)
 
 	err = connman_resolver_register(&dnsproxy_resolver);
 	if (err < 0)
-		destroy_listener();
+		goto destroy;
+
+	err = connman_notifier_register(&dnsproxy_notifier);
+	if (err < 0)
+		goto unregister;
+
+	return 0;
+
+unregister:
+	connman_resolver_unregister(&dnsproxy_resolver);
+
+destroy:
+	destroy_listener();
 
 	return err;
 }
 
 static void dnsproxy_exit(void)
 {
-	destroy_listener();
+	connman_notifier_unregister(&dnsproxy_notifier);
 
 	connman_resolver_unregister(&dnsproxy_resolver);
+
+	destroy_listener();
 }
 
 CONNMAN_PLUGIN_DEFINE(dnsproxy, "DNS proxy resolver plugin", VERSION,
