@@ -66,17 +66,25 @@ void __connman_agent_cleanup(void);
 int __connman_agent_register(const char *sender, const char *path);
 int __connman_agent_unregister(const char *sender, const char *path);
 
+struct connman_service;
+struct connman_ipconfig;
+
 int __connman_counter_register(const char *owner, const char *path,
 						unsigned int interval);
 int __connman_counter_unregister(const char *owner, const char *path);
 
-void __connman_counter_notify(const char *interface,
-				unsigned int rx_bytes, unsigned int tx_bytes);
+void __connman_counter_notify(struct connman_ipconfig *config,
+			unsigned int rx_packets, unsigned int tx_packets,
+			unsigned int rx_bytes, unsigned int tx_bytes,
+			unsigned int rx_error, unsigned int tx_error,
+			unsigned int rx_dropped, unsigned int tx_dropped);
+
+int __connman_counter_add_service(struct connman_service *service);
+void __connman_counter_remove_service(struct connman_service *service);
 
 int __connman_counter_init(void);
 void __connman_counter_cleanup(void);
 
-struct connman_service;
 
 typedef void (* passphrase_cb_t) (struct connman_service *service,
 				const char *passphrase, void *user_data);
@@ -253,6 +261,9 @@ int __connman_utsname_set_domainname(const char *domainname);
 
 #include <connman/timeserver.h>
 
+int __connman_timeserver_init(void);
+void __connman_timeserver_cleanup(void);
+
 #include <connman/dhcp.h>
 
 int __connman_dhcp_init(void);
@@ -271,10 +282,9 @@ void __connman_udev_start(void);
 void __connman_udev_cleanup(void);
 char *__connman_udev_get_devtype(const char *ifname);
 void __connman_udev_rfkill(const char *sysname, connman_bool_t blocked);
-void __connman_udev_enable_rfkill_processing(void);
 connman_bool_t __connman_udev_get_blocked(int phyindex);
 
-#include <connman/device.h>
+#include <connman/technology.h>
 
 void __connman_technology_list(DBusMessageIter *iter, void *user_data);
 
@@ -291,6 +301,15 @@ int __connman_technology_update_rfkill(unsigned int index,
 						connman_bool_t hardblock);
 int __connman_technology_remove_rfkill(unsigned int index);
 
+void __connman_technology_add_interface(enum connman_service_type type,
+				int index, const char *name, const char *ident);
+void __connman_technology_remove_interface(enum connman_service_type type,
+				int index, const char *name, const char *ident);
+int __connman_technology_enable_tethering(void);
+int __connman_technology_disable_tethering(void);
+
+#include <connman/device.h>
+
 int __connman_device_init(void);
 void __connman_device_cleanup(void);
 
@@ -382,6 +401,13 @@ int __connman_profile_add_network(struct connman_network *network);
 int __connman_profile_update_network(struct connman_network *network);
 int __connman_profile_remove_network(struct connman_network *network);
 
+int __connman_tethering_init(void);
+void __connman_tethering_cleanup(void);
+
+connman_bool_t __connman_tethering_get_status(void);
+int __connman_tethering_set_status(connman_bool_t status);
+void __connman_tethering_update_interface(const char *interface);
+
 #include <connman/service.h>
 
 int __connman_service_init(void);
@@ -417,6 +443,7 @@ int __connman_service_indicate_state(struct connman_service *service,
 int __connman_service_indicate_error(struct connman_service *service,
 					enum connman_service_error error);
 int __connman_service_indicate_default(struct connman_service *service);
+int __connman_service_request_login(struct connman_service *service);
 
 int __connman_service_lookup(const char *pattern, const char **path);
 int __connman_service_connect(struct connman_service *service);
@@ -431,6 +458,24 @@ void __connman_service_append_nameserver(struct connman_service *service,
 						const char *nameserver);
 void __connman_service_remove_nameserver(struct connman_service *service,
 						const char *nameserver);
+void __connman_service_nameserver_add_routes(struct connman_service *service,
+						const char *gw);
+void __connman_service_nameserver_del_routes(struct connman_service *service);
+
+unsigned long __connman_service_stats_get_rx_packets(struct connman_service *service);
+unsigned long __connman_service_stats_get_tx_packets(struct connman_service *service);
+unsigned long __connman_service_stats_get_rx_bytes(struct connman_service *service);
+unsigned long __connman_service_stats_get_tx_bytes(struct connman_service *service);
+unsigned long __connman_service_stats_get_rx_errors(struct connman_service *service);
+unsigned long __connman_service_stats_get_tx_errors(struct connman_service *service);
+unsigned long __connman_service_stats_get_rx_dropped(struct connman_service *service);
+unsigned long __connman_service_stats_get_tx_dropped(struct connman_service *service);
+unsigned long __connman_service_stats_get_time(struct connman_service *service);
+void __connman_service_stats_update(struct connman_service *service,
+				unsigned int rx_packets, unsigned int tx_packets,
+				unsigned int rx_bytes, unsigned int tx_bytes,
+				unsigned int rx_error, unsigned int tx_error,
+				unsigned int rx_dropped, unsigned int tx_dropped);
 
 #include <connman/location.h>
 
@@ -487,6 +532,8 @@ int __connman_rtnl_init(void);
 void __connman_rtnl_start(void);
 void __connman_rtnl_cleanup(void);
 
+unsigned int __connman_rtnl_update_interval_add(unsigned int interval);
+unsigned int __connman_rtnl_update_interval_remove(unsigned int interval);
 int __connman_rtnl_request_update(void);
 int __connman_rtnl_send(const void *buf, size_t len);
 
